use std::{
    fs::File,
    io::{Cursor, Read, Write},
    path::PathBuf,
};

use binrw::{BinRead, BinWriterExt};
use ezcad::{
    array_of::ArrayOf,
    layer::Layer,
    objects::{rectangle::Rectangle, Object},
    pen::Pen,
    types::Point,
};
use log::{debug, trace};
use rand::{seq::SliceRandom, thread_rng};
use serde::{Deserialize, Serialize};

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct ExportObject {
    layer: usize,
    object: usize,
    path: PathBuf,
}

impl ExportObject {
    pub fn export(&self, layers: &ArrayOf<Layer>) {
        debug!(
            "Exporting layer #{} object #{} to '{}'",
            self.layer,
            self.object,
            self.path.to_string_lossy()
        );

        let layer = layers.get(self.layer).expect("Invalid layer index");
        let object = layer
            .objects
            .get(self.object)
            .expect("Invalid object index");

        let mut buffer: Cursor<Vec<u8>> = Cursor::new(vec![]);
        buffer.write_le(object).expect("Failed to serialize object");

        let mut output: File = File::create(&self.path).expect("Failed to open output file");
        output
            .write_all(buffer.into_inner().as_slice())
            .expect("Failed to write to output file");
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct ImportObject {
    layer: usize,
    object: Option<usize>,
    path: PathBuf,
}

impl ImportObject {
    pub fn import(&self, layers: &mut ArrayOf<Layer>) {
        debug!(
            "Importing layer #{}{} from '{}'",
            self.layer,
            match &self.object {
                Some(object) => format!(" object #{}", object),
                None => String::new(),
            },
            self.path.to_string_lossy(),
        );

        let mut input: File = File::open(&self.path).expect("Failed to open input file");

        let mut buffer: Cursor<Vec<u8>> = Cursor::new(vec![]);
        input
            .read_to_end(buffer.get_mut())
            .expect("Failed to read input file");

        let object: Object =
            Object::read_le(&mut buffer).expect("Failed to deserialize input as object");

        let layer: &mut Layer = layers.get_mut(self.layer).expect("Invalid layer index");

        if let Some(index) = self.object {
            let dst: &mut Object = layer.objects.get_mut(index).expect("Invalid object index");
            *dst = object;
        } else {
            (*layer.objects).push(object);
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct DeleteObjects {
    layer: usize,
    object: Option<usize>,
}

impl DeleteObjects {
    pub fn delete(&self, layers: &mut ArrayOf<Layer>) {
        debug!(
            "Deleting {} from  layer #{}",
            match &self.object {
                Some(object) => format!("object #{}", object),
                None => format!("all objects"),
            },
            self.layer,
        );

        let layer: &mut Layer = layers.get_mut(self.layer).expect("Invalid layer index");

        match self.object {
            Some(index) => {
                assert!(index < layer.objects.len(), "Invalid object index");
                layer.objects.remove(index);
            }
            None => layer.objects.clear(),
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct RectangleArray {
    layer: usize,
    width: f64,
    height: f64,
    columns: usize,
    rows: usize,
    spacing: f64,
    z: f64,
    starting_pen: usize,
}

impl RectangleArray {
    pub fn generate(&self, pens: &Vec<Pen>, layers: &mut ArrayOf<Layer>) {
        debug!(
            "Generating {} x {} array of {} x {} rectangles with spacing of {} starting from pen #{} on layer #{}",
            self.columns, self.rows, self.width, self.height, self.spacing, self.starting_pen, self.layer
        );
        assert!(
            self.rows >= 1 && self.columns >= 1,
            "Invalid row/column value"
        );
        assert!(
            self.starting_pen + (self.rows * self.columns) < pens.len(),
            "Invalid starting pen"
        );

        let layer: &mut Layer = layers.get_mut(self.layer).expect("Invalid layer index");

        let incr_x: f64 = self.width + self.spacing;
        let incr_y: f64 = self.height + self.spacing;

        let bottom_left: Point = Point {
            x: (self.columns - 1) as f64 * -incr_x / 2.0,
            y: (self.rows - 1) as f64 * incr_y / 2.0,
        };

        // Closure that returns origin point of given index in array, where index starts
        // from top left and increments to the right and wraps around to the row below
        let calc_pt = |index: usize| {
            let x_pos: f64 = (index % self.columns) as f64;
            let y_pos: f64 = (index / self.columns) as f64;

            Point {
                x: bottom_left.x - incr_x * x_pos,
                y: bottom_left.y - incr_y * y_pos,
            }
        };

        // Randomize draw order
        let mut seq: Vec<usize> = (0..(self.rows * self.columns)).collect();
        seq.shuffle(&mut thread_rng());

        // Generate objects and append to layer
        for (pen_incr, obj_idx) in seq.into_iter().enumerate() {
            let mut rect: Rectangle = Rectangle::default();
            let origin: Point = calc_pt(obj_idx);

            *rect.corner_a = Point {
                x: origin.x - self.width / 2.0,
                y: origin.y - self.height / 2.0,
            };
            *rect.corner_b = Point {
                x: origin.x + self.width / 2.0,
                y: origin.y + self.height / 2.0,
            };
            *rect.core.origin = origin;
            *rect.core.pen = (self.starting_pen + pen_incr).try_into().unwrap();
            *rect.core.z = self.z;

            trace!(
                "Adding rectangle with pen #{} at {} (from {} to {})",
                *rect.core.pen,
                *rect.core.origin,
                *rect.corner_a,
                *rect.corner_b
            );

            layer.objects.push(Object::Rectangle(rect));
        }
    }
}
