use std::fmt::Debug;

use binrw::{BinRead, BinWrite};
use diff::Diff;
use modular_bitfield::{
    bitfield,
    specifiers::{B1, B14},
};

use crate::{
    field_of::FieldOf,
    types::{Field, ObjectType, Point, WString, F64, U16, U32},
};

use self::{
    circle::Circle, ellipse::Ellipse, hatch::Hatch, line::Lines, polygon::Polygon,
    rectangle::Rectangle,
};

pub mod circle;
pub mod ellipse;
pub mod hatch;
pub mod line;
pub mod polygon;
pub mod rectangle;

#[bitfield(bits = 16)]
#[derive(BinRead, BinWrite, Copy, Clone, Debug, Default, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
#[br(map = Self::from_bytes)]
#[bw(map = |&x| Self::into_bytes(x))]
pub struct ObjectFlags {
    pub disabled: B1,
    pub aspect_ratio_unlocked: B1,
    #[skip]
    __: B14,
}

/// Core properties defined for all object types
#[cfg_attr(feature = "default-debug", derive(Debug))]
#[derive(BinRead, BinWrite, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
#[brw(magic(17u32))]
pub struct ObjectCore {
    pub pen: U32,
    pub obj_type: FieldOf<ObjectType>,
    pub flags: FieldOf<ObjectFlags>,
    pub name: WString,
    pub count: U32,
    _unknown_1: Field,
    pub io_control_enable_mask: U16,
    pub io_control_disable_mask: U16,
    _unknown_2: [Field; 5],
    pub origin: FieldOf<Point>,
    pub z: F64,
    pub a: F64,
    _unknown_3: Field,
}

// Custom Debug implementation to only print known fields
#[cfg(not(feature = "default-debug"))]
impl Debug for ObjectCore {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("ObjectCore")
            .field("pen", &self.pen)
            .field("obj_type", &self.obj_type)
            .field("flags", &self.flags)
            .field("name", &self.name)
            .field("count", &self.count)
            .field("io_control_enable_mask", &self.io_control_enable_mask)
            .field("io_control_disable_mask", &self.io_control_disable_mask)
            .field("origin", &self.origin)
            .field("z", &self.z)
            .field("a", &self.a)
            .finish()
    }
}

impl Default for ObjectCore {
    fn default() -> Self {
        Self {
            pen: 0.into(),
            obj_type: ObjectType::Unknown.into(),
            flags: ObjectFlags::new()
                .with_disabled(0)
                .with_aspect_ratio_unlocked(1)
                .into(),
            name: String::new().into(),
            count: 0.into(),
            _unknown_1: vec![0, 0].into(), // 0_u16
            io_control_enable_mask: Default::default(),
            io_control_disable_mask: Default::default(),
            _unknown_2: [
                vec![0, 0].into(),                     // 0_u16
                vec![1, 0, 0, 0].into(),               // 1_u32
                vec![1, 0, 0, 0].into(),               // 1_u32
                vec![0, 0, 0, 0, 0, 0, 36, 64].into(), // 10.0_f64
                vec![0, 0, 0, 0, 0, 0, 36, 64].into(), // 10.0_f64
            ]
            .into(),
            origin: Point { x: 0.0, y: 0.0 }.into(),
            z: 0.0.into(),
            a: 0.0.into(),
            _unknown_3: vec![0, 0, 0, 0].into(), // 0_u32
        }
    }
}

#[derive(BinRead, BinWrite, Debug, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub enum Object {
    #[brw(magic = 1u32)]
    Curve(Lines),
    #[brw(magic = 2u32)]
    Point(Lines),
    #[brw(magic = 3u32)]
    Rectangle(Rectangle),
    #[brw(magic = 4u32)]
    Circle(Circle),
    #[brw(magic = 5u32)]
    Ellipse(Ellipse),
    #[brw(magic = 6u32)]
    Polygon(Polygon),
    #[brw(magic = 32u32)]
    Hatch(Hatch),
}
